


#ifndef _TAPIDEMO_H
#define _TAPIDEMO_H
/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************
   Module      : tapidemo.h
   Description  :
 ****************************************************************************
   \file tapidemo.h

   \remarks

   \note Changes:
*******************************************************************************/

/* ============================= */
/* Includes                      */
/* ============================= */

#include "common.h"

#ifdef ABACUS_TEST
#include "./abacus/com_client.h"
#include "./abacus/dev_help.h"
#endif /* ABACUS_TEST */

#ifdef ABACUS_TEST
extern IFX_int32_t nVerifySystemInitializtion;
extern CTRL_STATUS_t* pCtrlStatus;
#ifdef EASY336
/** PCM configuration */
extern RM_SVIP_PCM_CFG_t pcm_cfg;
#endif /* EASY336 */
#endif /* ABACUS_TEST */

/* TESTING - flag used when tapidemo starts on different versions. Moved
   to makefile.am, in VxWorks add this flag to compiler options */
/*#define STREAM_1_1*/ /* Default is version 1.2 - common voice api */
/* If flag above is not set version 1.1 is used */

#ifdef AM5120
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#endif


/* ============================= */
/* Global Defines                */
/* ============================= */

/* Tapi version  4 takes those enums from lib_vos */
#ifndef TAPI_VERSION4
#ifdef VXWORKS

#elif (defined (LINUX))

enum
{
   ERROR = -1,
   OK    = 0
};

enum
{
   WAIT_FOREVER = -1,
   NO_WAIT = 0
};

#endif /* VXWORKS || LINUX */
#endif /* TAPI_VERSION4 */

#ifdef DEBUG
#define deb(msg) printf msg;
#else
#define deb(msg)
#endif

/** time out for select in seconds */
#define  SELECT_TIMEOUT_SEC      0
#ifndef VXWORKS
/** time out for select in useconds (50000us=50ms=0.05s) */
#define  SELECT_TIMEOUT_USEC     50000
#else /* VXWORKS */
/** time out for select in useconds (500us=0.50ms=0.0005s),
    for VXWORKS select doesn't reset timeout structure
    and if timeout is too small FXO will not be checked for ringing timeout */
#define  SELECT_TIMEOUT_USEC     500
#endif /* VXWORKS */

/* ============================= */
/* Local Macros & Definitions    */
/* ============================= */

/** max length of strings in PrintCallProgress */
#define PCP_MAX_STR_LEN 80

/* ============================= */
/* Global Structures and enums   */
/* ============================= */

/** IDs for board combination */
enum _BOARD_COMB_ID
{
   ONE_CPE = 0,
   ONE_VIP = 1,
   ONE_VMMC = 2,
   ONE_VMMC_ONE_CPE = 3,
   ONE_DXT = 4,
   TWO_DXTS = 5,
   ONE_VMMC_VINAX = 6,
   ONE_VMMC_EASY508XX = 7,
   X_SVIP = 8,
   X_SVIP_X_XT16 = 9,
   X_XT16 = 10,
   ONE_EASY50712_ONE_TERIDIAN73M1866,
   ONE_EASY508XX_ONE_TERIDIAN73M1866,
   ONE_EASY80800_ONE_TERIDIAN73M1866,
   ONE_EASY80910,
   UNSUPPORTED_COMB
};

/** Max user name length. */
enum { MAX_USERNAME_LEN = 20 };

/** Time which must pass for ONHOOK to be valid over FXO line [in seconds]. */
enum { TIMEOUT_FOR_ONHOOK = 5 };

/** Time which must pass for waiting on answer from second phone
   during external call */ 
enum { TIMEOUT_FOR_EXTERNAL_CALL = 5 };

/** Default Voice Activity Detection (VAD) configuration */
enum { DEFAULT_VAD_CFG = IFX_TAPI_ENC_VAD_NOVAD };

/* Holding board combination names. */
extern const IFX_char_t* BOARD_COMB_NAMES[MAX_COMBINATIONS];

/* ============================= */
/* Global variable definition    */
/* ============================= */

/** Holding trace redirection settings */
extern TRACE_REDIRECTION_CTRL_t oTraceRedirection;

/* ============================= */
/* Global function declaration   */
/* ============================= */

extern IFX_return_t TAPIDEMO_SetAction(CTRL_STATUS_t* pCtrl,
                                       PHONE_t* pPhone,
                                       CONNECTION_t* pConn,
                                       IFX_int32_t nAction);
#ifdef ABACUS_TEST
extern IFX_int32_t TAPIDEMO_InitAdminSocket(IFX_void_t);
#endif /* ABACUS_TEST */
extern IFX_int32_t TAPIDEMO_SetDefaultAddr(PROGRAM_ARG_t* pProgramArg);

extern IFX_void_t TAPIDEMO_ClearPhone(CTRL_STATUS_t* pCtrl,
                                      PHONE_t* pPhone,
                                      CONNECTION_t* pConn,
                                      INTERNAL_EVENTS_t nInternalEvent,
                                      BOARD_t* pBoard);
#endif /* _SYSTEM_H */

