/******************************************************************************

                               Copyright (c) 2008
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

  ****************************************************************************/


/**
   \file voip.c
   \date 2006-04-04
   \brief Working with coder and signal module.

   This file implements methods to handle with data channel resources, channel
   mapping, start/stop coder/encoder, hanling data packets from dev or socket,
   initialization, ...
*/

/* ============================= */
/* Includes                      */
/* ============================= */

#include "tapidemo.h"
#include "cid.h"
#include "voip.h"

#ifdef EASY336
#include "svip_io.h"
#include "board_easy336.h"
#include "lib_svip.h"
#include "lib_svip_rm.h"
#include "common.h"
#include "abstract.h"
#endif /* EASY336 */

/* ============================= */
/* Global Defines                */
/* ============================= */

#ifdef EASY336
#define IFX_SVIP_SYS_COD_NUM  (1*16)
#define IFX_SVIP_UDP_END      ((IFX_SVIP_UDP_START) + (IFX_SVIP_SYS_COD_NUM))
/** VoFW destination UDP port number. */
#define FW_PORT     6000
#endif /* EASY336 */

/* ============================= */
/* Global Structures             */
/* ============================= */

/* ============================= */
/* Global function declaration   */
/* ============================= */

/* ============================= */
/* Local variable definition     */
/* ============================= */

/** Array of pointers to strings with codecs names. */
IFX_char_t* pCodecName[IFX_TAPI_ENC_TYPE_MAX];
/** Array to convert enums IFX_TAPI_COD_LENGTH_t
    to coresponding values names. */
IFX_char_t* oFrameLen[FRAME_SIZE_COUNT];

/* --------------------------------------------------------- */
/*             RTP Payload Type Tables START                 */
/* --------------------------------------------------------- */

/** Payload types table - Upstream */
enum
{
   RTP_ENC_TYPE_MLAW_UP = 0x00,
   RTP_ENC_TYPE_ALAW_UP = 0x08,
   RTP_ENC_TYPE_G729_UP = 0x12,
   RTP_ENC_TYPE_G723_53_UP = 0x04,
   RTP_ENC_TYPE_G723_63_UP = 0x04,
#ifndef STREAM_1_1
   RTP_ENC_TYPE_G7221_24_UP = 0x73,
   RTP_ENC_TYPE_G7221_32_UP = 0x74,
   RTP_ENC_TYPE_G722_64_UP = 0x09,
#endif /* STREAM_1_1 */
   RTP_ENC_TYPE_G728_UP = 0x0F,
   RTP_ENC_TYPE_G729_E_UP = 0x61,
   RTP_ENC_TYPE_G726_16_UP = 0x63,
   RTP_ENC_TYPE_G726_24_UP = 0x64,
   RTP_ENC_TYPE_G726_32_UP = 0x65,
   RTP_ENC_TYPE_G726_40_UP = 0x66,
   RTP_ENC_TYPE_ILBC_133_UP = 0x67,
   RTP_ENC_TYPE_ILBC_152_UP = 0x67,
   RTP_ENC_TYPE_LIN16_16_UP = 0x70,
   RTP_ENC_TYPE_LIN16_8_UP = 0x71,
   RTP_ENC_TYPE_AMR_4_75_UP = 0x72,
   RTP_ENC_TYPE_AMR_5_9_UP = 0x72,
   RTP_ENC_TYPE_AMR_5_15_UP = 0x72,
   RTP_ENC_TYPE_AMR_6_7_UP = 0x72,
   RTP_ENC_TYPE_AMR_7_4_UP = 0x72,
   RTP_ENC_TYPE_AMR_7_95_UP = 0x72,
   RTP_ENC_TYPE_AMR_10_2_UP = 0x72,
   RTP_ENC_TYPE_AMR_12_2_UP = 0x72,
   RTP_ENC_TYPE_MLAW_VBD_UP = 0x75,
   RTP_ENC_TYPE_ALAW_VBD_UP = 0x76

};

/** Payload types table - Downstream */
enum
{
   RTP_ENC_TYPE_MLAW_DOWN = 0x00,
   RTP_ENC_TYPE_ALAW_DOWN = 0x08,
   RTP_ENC_TYPE_G729_DOWN = 0x12,
   RTP_ENC_TYPE_G723_53_DOWN = 0x04,
   RTP_ENC_TYPE_G723_63_DOWN = 0x04,
#ifndef STREAM_1_1
   RTP_ENC_TYPE_G7221_24_DOWN = 0x73,
   RTP_ENC_TYPE_G7221_32_DOWN = 0x74,
   RTP_ENC_TYPE_G722_64_DOWN = 0x09,
#endif /* STREAM_1_1 */
   RTP_ENC_TYPE_G728_DOWN = 0x0F,
   RTP_ENC_TYPE_G729_E_DOWN = 0x61,
   RTP_ENC_TYPE_G726_16_DOWN = 0x63,
   RTP_ENC_TYPE_G726_24_DOWN = 0x64,
   RTP_ENC_TYPE_G726_32_DOWN = 0x65,
   RTP_ENC_TYPE_G726_40_DOWN = 0x66,
   RTP_ENC_TYPE_ILBC_133_DOWN = 0x67,
   RTP_ENC_TYPE_ILBC_152_DOWN = 0x67,
   RTP_ENC_TYPE_LIN16_16_DOWN = 0x70,
   RTP_ENC_TYPE_LIN16_8_DOWN = 0x71,
   RTP_ENC_TYPE_AMR_4_75_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_5_9_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_5_15_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_6_7_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_7_4_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_7_95_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_10_2_DOWN = 0x72,
   RTP_ENC_TYPE_AMR_12_2_DOWN = 0x72,
   RTP_ENC_TYPE_MLAW_VBD_DOWN = 0x75,
   RTP_ENC_TYPE_ALAW_VBD_DOWN = 0x76
};

/* --------------------------------------------------------- */
/*               RTP Payload Type Tables END                 */
/* --------------------------------------------------------- */

/** Jitter buffer configuration. */

/** Packet adaptation */
static const IFX_int32_t JB_PACKET_ADAPTATION = IFX_TAPI_JB_PKT_ADAPT_VOICE;
/** Local adaptation */
static const IFX_int32_t JB_LOCAL_ADAPTATION  = IFX_TAPI_JB_LOCAL_ADAPT_OFF;
/** Play out dealy, value between 0..16 */
static const IFX_int32_t JB_SCALING    = 0x16;
/** Initial size of jitter buffer in timestamps of 125 us */
static const IFX_int32_t JB_INITIAL_SIZE      = 0x0050;
/** Maximum size of jitter buffer in timestamps of 125 us */
static const IFX_int32_t JB_MAXIMUM_SIZE      = 0x05A0;
/** Minimum size of jitter buffer in timestamps of 125 us */
static const IFX_int32_t JB_MINIMUM_SIZE      = 0x0050;

/** Default encoder type */
enum { DEFAULT_CODER_TYPE = IFX_TAPI_ENC_TYPE_MLAW };

/** Default packetisation time [ms] */
/*enum { DEFAULT_PACKETISATION_TIME = 10 };*/

/** Packetisation time is no longer given in ms
    but described by enum value from IFX_TAPI_COD_LENGTH_t */
enum { DEFAULT_PACKETISATION_TIME = IFX_TAPI_COD_LENGTH_10 };

/** Structure holding codecs with default packetisation length  */
typedef struct _CODEC_LEN_t
{
   IFX_TAPI_ENC_TYPE_t nCodecName;
   IFX_TAPI_COD_LENGTH_t nDefaultLenght;
} CODEC_LEN_t;

/** List of combinations of codecs with default packetisation length */
static const CODEC_LEN_t CODEC_LEN[] =
{
   { IFX_TAPI_ENC_TYPE_G723_63, IFX_TAPI_COD_LENGTH_30 },
   { IFX_TAPI_ENC_TYPE_G723_53, IFX_TAPI_COD_LENGTH_30 },
   /* Codec below doesn't work */
   { IFX_TAPI_ENC_TYPE_G728, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G729, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_MLAW, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_ALAW, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_MLAW_VBD, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_ALAW_VBD, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G726_16, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G726_24, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G726_32, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G726_40, IFX_TAPI_COD_LENGTH_10 },
   /* Codec below doesn't work */
   { IFX_TAPI_ENC_TYPE_G729_E, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_ILBC_133, IFX_TAPI_COD_LENGTH_30 },
   { IFX_TAPI_ENC_TYPE_ILBC_152, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_LIN16_8, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_LIN16_16, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_AMR_4_75, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_5_15, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_5_9, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_6_7, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_7_4, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_7_95, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_10_2, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_AMR_12_2, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_G722_64, IFX_TAPI_COD_LENGTH_10 },
   { IFX_TAPI_ENC_TYPE_G7221_24, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_G7221_32, IFX_TAPI_COD_LENGTH_20 },
   { IFX_TAPI_ENC_TYPE_MAX, IFX_TAPI_COD_LENGTH_ZERO }
};

/** Current encoder type to use */
IFX_TAPI_ENC_TYPE_t eCurrEncType = DEFAULT_CODER_TYPE;

/** Current packetisation time to use */
static IFX_int32_t nCurrPacketTime = DEFAULT_PACKETISATION_TIME;

/** Number of free socket available */
static IFX_int32_t nFreeSocketPortAvail = 0;

/* ============================= */
/* Local function declaration    */
/* ============================= */

/* ============================= */
/* Local function definition     */
/* ============================= */

/**
   Function to set array of pointers to codec names.

   \return IFX_void_t
*/
IFX_void_t VOIP_SetCodecNames(IFX_void_t)
{
   IFX_int32_t i;

   for (i=0; i<IFX_TAPI_ENC_TYPE_MAX; i++)
   {
      /* set begining value */
      pCodecName[i] = "NO CODEC";
   }

   /* point to strings with codecs names */
   pCodecName[IFX_TAPI_ENC_TYPE_G723_63] = "G.723.1 with 6.3kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G723_53] = "G.723.1 with 5.3kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G728] = "G.728 with 16kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G729] = "G.729A with 8kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_MLAW] = "G.711 u-Law with 64kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_ALAW] = "G.711 A-Law with 64kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_MLAW_VBD] = "G.711 u-Law VBD with 64kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_ALAW_VBD] = "G.711 A-Law VBD with 64kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G726_16] = "G.726 with 16kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G726_24] = "G.726 with 24kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G726_32] = "G.726 with 32kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G726_40] = "G.726 with 40kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G729_E] = "G.729E with 11.8kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_ILBC_133] = "iLBC with 13.3kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_ILBC_152] = "iLBC with 15.2kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_LIN16_8] = "Linear Codec with 8kHz";
   pCodecName[IFX_TAPI_ENC_TYPE_LIN16_16] = "Linear Codec with 16kHz";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_4_75] = "AMR with 4.75kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_5_15] = "AMR with 5.15kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_5_9] = "AMR with 5.9kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_6_7] = "AMR with 6.7kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_7_4] = "AMR with 7.4kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_7_95] = "AMR with 7.95kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_10_2] = "AMR with 10.2kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_AMR_12_2] = "AMR with 12.2kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G722_64] = "G.722 with 64kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G7221_24] = "G.722.1 with 24kbit/s";
   pCodecName[IFX_TAPI_ENC_TYPE_G7221_32] = "G.722.1 with  32kbit/s";

   /* set frames length in ms */
   oFrameLen[IFX_TAPI_COD_LENGTH_ZERO]="0ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_2_5]="2.5ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_5]="5ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_5_5]="5.5ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_10]="10ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_11]="11ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_20]="20ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_30]="30ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_40]="40ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_50]="50ms";
   oFrameLen[IFX_TAPI_COD_LENGTH_60]="60ms";

   return;
}

/* ============================= */
/* Global function definition    */
/* ============================= */


/**
   VoIP Init.

   \param pProgramArg  - pointer to program arguments
   \param pBoard - board handle

   \return IFX_SUCCESS on OK, otherwise IFX_ERROR
*/
IFX_return_t VOIP_Init(PROGRAM_ARG_t* pProgramArg, BOARD_t* pBoard)
{
   IFX_int32_t i = 0;
   IFX_TAPI_PKT_RTP_PT_CFG_t rtpPTConf;
   IFX_TAPI_JB_CFG_t jbCfgVoice;
   IFX_TAPI_PKT_RTP_CFG_t rtpConf;
   IFX_return_t ret;
#ifdef TAPI_VERSION4
   IFX_TAPI_ENC_VAD_CFG_t oVadCfg;
#endif /* TAPI_VERSION4 */
   IFX_TAPI_ENC_TYPE_t nEncTypeTmp = eCurrEncType;
   IFX_int32_t nPacketTimeTmp = nCurrPacketTime;

#ifndef STREAM_1_1
   IFX_int32_t fd_ch = -1;
#endif /* STREAM_1_1 */

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pProgramArg, IFX_ERROR);
   if (0 >= pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Err, VOIP_Init: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_ERROR;
   }

   /* check if non default codec is supported */
   if (DEFAULT_CODER_TYPE != nEncTypeTmp)
   {
      /* check if codec suported */
      if (IFX_SUCCESS != VOIP_CodecAvailabilityCheck(pBoard,
                                                     nEncTypeTmp))
      {
         nEncTypeTmp = DEFAULT_CODER_TYPE;
         nPacketTimeTmp = DEFAULT_PACKETISATION_TIME;
         TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
               ("%s: Using default codec %s, %s.\n",
                pBoard->pszBoardName,
                pCodecName[DEFAULT_CODER_TYPE],
                oFrameLen[DEFAULT_PACKETISATION_TIME]));
      }
   }

   memset(&rtpPTConf, 0, sizeof(IFX_TAPI_PKT_RTP_PT_CFG_t));

   /* Payload types table - Upstream */

   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_MLAW]       = RTP_ENC_TYPE_MLAW_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_ALAW]       = RTP_ENC_TYPE_ALAW_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G729]       = RTP_ENC_TYPE_G729_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G723_53]    = RTP_ENC_TYPE_G723_53_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G723_63]    = RTP_ENC_TYPE_G723_63_UP;
#ifndef STREAM_1_1
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G7221_24]   = RTP_ENC_TYPE_G7221_24_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G7221_32]   = RTP_ENC_TYPE_G7221_32_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G722_64]    = RTP_ENC_TYPE_G722_64_UP;
#endif /* STREAM_1_1 */
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G726_16]    = RTP_ENC_TYPE_G726_16_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G726_24]    = RTP_ENC_TYPE_G726_24_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G726_32]    = RTP_ENC_TYPE_G726_32_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G726_40]    = RTP_ENC_TYPE_G726_40_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G728]       = RTP_ENC_TYPE_G728_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_G729_E]     = RTP_ENC_TYPE_G729_E_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_ILBC_133]   = RTP_ENC_TYPE_ILBC_133_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_ILBC_152]   = RTP_ENC_TYPE_ILBC_152_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_LIN16_16]   = RTP_ENC_TYPE_LIN16_16_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_LIN16_8]    = RTP_ENC_TYPE_LIN16_8_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_4_75]   = RTP_ENC_TYPE_AMR_4_75_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_5_9]    = RTP_ENC_TYPE_AMR_5_9_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_5_15]   = RTP_ENC_TYPE_AMR_5_15_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_6_7]    = RTP_ENC_TYPE_AMR_6_7_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_7_4]    = RTP_ENC_TYPE_AMR_7_4_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_7_95]   = RTP_ENC_TYPE_AMR_7_95_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_10_2]   = RTP_ENC_TYPE_AMR_10_2_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_AMR_12_2]   = RTP_ENC_TYPE_AMR_12_2_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_MLAW_VBD]   = RTP_ENC_TYPE_MLAW_VBD_UP;
   rtpPTConf.nPTup[IFX_TAPI_ENC_TYPE_ALAW_VBD]   = RTP_ENC_TYPE_ALAW_VBD_UP;

   /* Payload types table - Downstream */

   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_MLAW]     = RTP_ENC_TYPE_MLAW_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_ALAW]     = RTP_ENC_TYPE_ALAW_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G729]     = RTP_ENC_TYPE_G729_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G723_53]  = RTP_ENC_TYPE_G723_53_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G723_63]  = RTP_ENC_TYPE_G723_63_DOWN;
#ifndef STREAM_1_1
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G7221_24] = RTP_ENC_TYPE_G7221_24_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G7221_32] = RTP_ENC_TYPE_G7221_32_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G722_64]  = RTP_ENC_TYPE_G722_64_DOWN;
#endif /* STREAM_1_1 */
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G726_16]  = RTP_ENC_TYPE_G726_16_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G726_24]  = RTP_ENC_TYPE_G726_24_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G726_32]  = RTP_ENC_TYPE_G726_32_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G726_40]  = RTP_ENC_TYPE_G726_40_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G728]     = RTP_ENC_TYPE_G728_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_G729_E]   = RTP_ENC_TYPE_G729_E_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_ILBC_133] = RTP_ENC_TYPE_ILBC_133_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_ILBC_152] = RTP_ENC_TYPE_ILBC_152_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_LIN16_16] = RTP_ENC_TYPE_LIN16_16_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_LIN16_8]  = RTP_ENC_TYPE_LIN16_8_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_4_75] = RTP_ENC_TYPE_AMR_4_75_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_5_9]  = RTP_ENC_TYPE_AMR_5_9_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_5_15] = RTP_ENC_TYPE_AMR_5_15_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_6_7]  = RTP_ENC_TYPE_AMR_6_7_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_7_4]  = RTP_ENC_TYPE_AMR_7_4_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_7_95] = RTP_ENC_TYPE_AMR_7_95_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_10_2] = RTP_ENC_TYPE_AMR_10_2_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_AMR_12_2] = RTP_ENC_TYPE_AMR_12_2_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_MLAW_VBD] = RTP_ENC_TYPE_MLAW_VBD_DOWN;
   rtpPTConf.nPTdown[IFX_TAPI_ENC_TYPE_ALAW_VBD] = RTP_ENC_TYPE_ALAW_VBD_DOWN;

   memset(&jbCfgVoice, 0, sizeof(IFX_TAPI_JB_CFG_t));

   /* Adaptive JB */
   jbCfgVoice.nJbType = IFX_TAPI_JB_TYPE_ADAPTIVE;
   /* Optimization for voice */
   jbCfgVoice.nPckAdpt = JB_PACKET_ADAPTATION;
   jbCfgVoice.nLocalAdpt = JB_LOCAL_ADAPTATION;
   jbCfgVoice.nScaling = JB_SCALING;
   /* Initial JB size 10 ms = 0x0050 * 125 us */
   jbCfgVoice.nInitialSize = JB_INITIAL_SIZE;
   /* Minimum JB size 10 ms = 0x0050 * 125 us */
   jbCfgVoice.nMinSize = JB_MINIMUM_SIZE;
   /* Maximum JB size 180 ms = 0x05A0 * 125 us */
   jbCfgVoice.nMaxSize = JB_MAXIMUM_SIZE;

#ifndef TAPI_VERSION4
   /* Set default RTP and RTCP settings */
   /* Start value for the sequence number. */
   pBoard->rtpConf.nSeqNr = 0;

   /* Synchronization source value for the voice and SID packets.
      Note: A change in the SSRC leads to a reset of the RTCP statistics.*/
   pBoard->rtpConf.nSsrc = 0;
   /* Defines how to handle RFC 2833 packets in upstream direction. Set
      parameter as defined in \ref IFX_TAPI_PKT_EV_OOB_t enumerator. */
   pBoard->rtpConf.nEvents = IFX_TAPI_PKT_EV_OOB_NO;

   /*  Defines whether the received RFC 2833 packets have to be played out.
       Set parameter as defined in \ref IFX_TAPI_PKT_EV_OOBPLAY_t enumerator. */
   pBoard->rtpConf.nPlayEvents = IFX_TAPI_PKT_EV_OOBPLAY_PLAY;

   /*  Payload type to be used for RFC 2833 frames in encoder direction
      (upstream). */
   pBoard->rtpConf.nEventPT = OOB_PAYLOAD;

   rtpConf = pBoard->rtpConf;
#endif /* TAPI_VERSION4 */

   for (i = 0; i < pBoard->nMaxCoderCh; i++)
   {
      pBoard->pDataChStat[i].fMapped = IFX_FALSE;
      pBoard->pDataChStat[i].fCodecStarted = IFX_FALSE;
      pBoard->pDataChStat[i].fStartCodec = IFX_TRUE;
      pBoard->pDataChStat[i].fDefaultVad = IFX_TRUE;
      pBoard->pDataChStat[i].nCodecType = nEncTypeTmp;
      pBoard->pDataChStat[i].nFrameLen = nPacketTimeTmp;
      pBoard->pDataChStat[i].nBitPack = IFX_TAPI_COD_RTP_BITPACK;
      pBoard->pDataChStat[i].nPrevCodecType = nEncTypeTmp;
      pBoard->pDataChStat[i].nPrevFrameLen = nPacketTimeTmp;
      pBoard->pDataChStat[i].fRestoreChannel = IFX_FALSE;

      memset(&pBoard->pDataChStat[i].oPrevLECcfg, 0, \
             sizeof(IFX_TAPI_WLEC_CFG_t));
      pBoard->pDataChStat[i].oPrevLECcfg.nType = IFX_TAPI_WLEC_TYPE_NFE;
      pBoard->pDataChStat[i].oPrevLECcfg.bNlp = IFX_TAPI_WLEC_NLP_ON;

#ifndef STREAM_1_1
      memset(&pBoard->pDataChStat[i].tapi_jb_PrevConf, 0, \
             sizeof(IFX_TAPI_JB_CFG_t));
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nJbType = 
         IFX_TAPI_JB_TYPE_ADAPTIVE;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nPckAdpt = JB_PACKET_ADAPTATION;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nLocalAdpt = JB_LOCAL_ADAPTATION;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nScaling  = JB_SCALING;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nInitialSize = JB_INITIAL_SIZE;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nMaxSize = JB_MAXIMUM_SIZE;
      pBoard->pDataChStat[i].tapi_jb_PrevConf.nMinSize = JB_MINIMUM_SIZE;

      fd_ch = VOIP_GetFD_OfCh(i, pBoard);

      /* Program the channel (addressed by fd) with
         the specified payload types. */
#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
         rtpPTConf.dev = pBoard->pDataChStat[i].nDev;
         rtpPTConf.ch = pBoard->pDataChStat[i].nCh;
         TD_ASSERT ((ret = ioctl(fd_ch,
            IFX_TAPI_PKT_RTP_PT_CFG_SET, (IFX_int32_t) &rtpPTConf)) ==
            IFX_SUCCESS, fd_ch, pBoard, rtpPTConf.dev,
            rtpPTConf.ch);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("IFX_TAPI_PKT_RTP_PT_CFG_SET failed for device %d channel %d, "
               "(File: %s, line: %d)\n",
               rtpPTConf.dev, rtpPTConf.ch, __FILE__, __LINE__));
            return ret;
         }
         /* Set jitter buffer */
         jbCfgVoice.dev = pBoard->pDataChStat[i].nDev;
         jbCfgVoice.ch = pBoard->pDataChStat[i].nCh;
         TD_ASSERT ((ret = ioctl(fd_ch, IFX_TAPI_JB_CFG_SET,
            (IFX_int32_t) &jbCfgVoice)) == IFX_SUCCESS, fd_ch,
            pBoard, jbCfgVoice.dev, jbCfgVoice.ch);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_JB_CFG_SET failed for device %d channel %d, "
                   "(File: %s, line: %d)\n",
                   jbCfgVoice.dev, jbCfgVoice.ch, __FILE__, __LINE__));
            return ret;
         }

         /* Voice Activity Detection (VAD) configuration */
         oVadCfg.dev =pBoard->pDataChStat[i].nDev;
         oVadCfg.ch =pBoard->pDataChStat[i].nCh;
         oVadCfg.vadMode = pProgramArg->nVadCfg;
         TD_ASSERT ((ret = ioctl(fd_ch, IFX_TAPI_ENC_VAD_CFG_SET,
            (IFX_int32_t) &oVadCfg)) == IFX_SUCCESS, fd_ch,
            pBoard, oVadCfg.dev, oVadCfg.ch);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_ENC_VAD_CFG_SET failed for device %d channel %d, "
                   "(File: %s, line: %d)\n",
                   oVadCfg.dev, oVadCfg.ch, __FILE__, __LINE__));
            return ret;
         }
      } /* if (pBoard->fSingleFD */
      else
#endif /* TAPI_VERSION4 */
      {
         /* set RTP payload type table */
         ret = ioctl(fd_ch, IFX_TAPI_PKT_RTP_PT_CFG_SET,
                     (IFX_int32_t) &rtpPTConf);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_PKT_RTP_PT_CFG_SET failed for channel %d, fd %d, "
                   "(File: %s, line: %d)\n",
                   i, fd_ch, __FILE__, __LINE__));
         }

         /* Set jitter buffer */
         ret = ioctl(fd_ch, IFX_TAPI_JB_CFG_SET, (IFX_int32_t) &jbCfgVoice);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_JB_CFG_SET failed for channel %d, fd %d, "
                   "(File: %s, line: %d)\n",
                   i, fd_ch, __FILE__, __LINE__));
         }

         /* Set RTP and RTCP */
         ret =  ioctl(fd_ch, IFX_TAPI_PKT_RTP_CFG_SET, (IFX_int32_t) &rtpConf);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_PKT_RTP_CFG_SET failed for channel %d, fd %d, "
                   "(File: %s, line: %d)\n",
                   i, fd_ch, __FILE__, __LINE__));
            Common_ErrorHandle (pBoard->nDevCtrl_FD, pBoard, 0);
         }

         /* Voice Activity Detection (VAD) configuration */
         ret =  ioctl(fd_ch, IFX_TAPI_ENC_VAD_CFG_SET,
                      (IFX_int32_t) pProgramArg->nVadCfg);
         if (ret != IFX_SUCCESS)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                  ("IFX_TAPI_ENC_VAD_CFG_SET failed to set %d "
                   "for channel %d, fd %d, (File: %s, line: %d)\n",
                   pProgramArg->nVadCfg, i, fd_ch, __FILE__, __LINE__));
         }
      }
#endif /* STREAM_1_1 */

#ifdef EASY336
      SVIP_NAT_LibInit();
#endif /* EASY336 */

   } /* for */

#ifdef TAPI_VERSION4
   if (pBoard->fUseSockets)
#endif
   {
      nFreeSocketPortAvail += pBoard->nMaxCoderCh;
   }
   /* set default codec and packetisation in PROGRAM_ARG_t structure */
   if (0 == pProgramArg->nEnCoderType)
   {
      pProgramArg->nEnCoderType = eCurrEncType;
   }
   if (0 == pProgramArg->nPacketisationTime)
   {
      pProgramArg->nPacketisationTime = nCurrPacketTime;
   }
   return IFX_SUCCESS;
} /* VOIP_Init() */


/**
   VoIP release.

   \param pProgramArg  - pointer to program arguments
   \param pBoard - board handle

   \return IFX_SUCCESS on OK, otherwise IFX_ERROR
*/
IFX_return_t VOIP_Release(PROGRAM_ARG_t* pProgramArg, BOARD_t* pBoard)
{
   return IFX_SUCCESS;
} /* VOIP_Release() */


/**
   Set VOIP encoder, vocoder type.

   \param nEncoderType - new encoder type

   \return IFX_SUCCESS if OK, IFX_ERROR if errors
   \remark this function sets only eCurrEncType,
           eCurrEncType is used during VOIP_Init() to set codec, this function
           shouldn't be used after VOIP_Init()
*/
IFX_return_t VOIP_SetEncoderType(IFX_int32_t nEncoderType)
{
   IFX_TAPI_ENC_TYPE_t encoder_type = DEFAULT_CODER_TYPE;

   memset(&encoder_type, 0, sizeof(encoder_type));

   if ((0 > nEncoderType) || (IFX_TAPI_ENC_TYPE_MAX < nEncoderType))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s), will use default"
            " DEFAULT_CODER_TYPE encoder type. "
            "(File: %s, line: %d)\n", __FILE__, __LINE__));
      /* Don't send error, but use default value. */
      encoder_type = DEFAULT_CODER_TYPE;
   }
   else
   {
      encoder_type = (IFX_TAPI_ENC_TYPE_t) (nEncoderType);
   }

   eCurrEncType = encoder_type;

   return IFX_SUCCESS;
} /* VOIP_SetEncoderType() */


/**
   Set packetization time to be used.

   \param nFrameLen - number of frame len (packetisation time) enum

   \return IFX_SUCCESS if OK, IFX_ERROR if errors
   \remark this function sets only nCurrPacketTime according to eCurrEncType,
           eCurrEncType is used during VOIP_Init() to set codec, this function
           shouldn't be used after VOIP_Init()
*/
IFX_return_t VOIP_SetFrameLen(IFX_int32_t nFrameLen)
{
   if ((0 >= nFrameLen) || (FRAME_SIZE_COUNT < nFrameLen))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument, use default packetisation time."
            " (File: %s, line: %d)\n", __FILE__, __LINE__));

      /* Don't return error, but use default value. */
      VOIP_SetDefaultFrameLen();
   }
   else
   {
      nCurrPacketTime = nFrameLen;
      /* Print encoder type with packetisation */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Set used encoder type %s\n"
            "    with frame length (packetisation time) %s\n",
            pCodecName[eCurrEncType], oFrameLen[nCurrPacketTime]));
   }

   return IFX_SUCCESS;
} /* VOIP_SetFrameLen() */

/**
   Set default packetization time.

   \return IFX_SUCCESS if OK, IFX_ERROR if errors
   \remark this function sets only nCurrPacketTime according to eCurrEncType,
           eCurrEncType is used during VOIP_Init() to set codec, this function
           shouldn't be used after VOIP_Init()
*/
IFX_return_t VOIP_SetDefaultFrameLen(IFX_void_t)
{
   IFX_int32_t frame_len = IFX_ERROR;

   /* get default frame length */
   frame_len = VOIP_GetCodecDefaultFrameLen(eCurrEncType);

   if ((0 >= frame_len) || (FRAME_SIZE_COUNT < frame_len))
   {
      /* Don't return error, but use default value. */
      frame_len = DEFAULT_PACKETISATION_TIME;
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Err, Can not find default frame length for encoder type %s,\n"
            "     (File: %s, line: %d)\n",
            pCodecName[eCurrEncType], __FILE__, __LINE__));
   }

   nCurrPacketTime = frame_len;
   /* Print encoder type with packetisation */
   TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
        ("Set used encoder type %s\n"
         "    with frame length (packetisation time) %s\n",
         pCodecName[eCurrEncType], oFrameLen[nCurrPacketTime]));

   return IFX_SUCCESS;
} /* VOIP_SetDefaultFrameLen() */

/**
   Get default packetization time.

   \param nCodec  - number of codec

   \return lowest supported packetization size for selected codec,
           IFX_ERROR if not found
*/
IFX_int32_t VOIP_GetCodecDefaultFrameLen(IFX_TAPI_ENC_TYPE_t nCodec)
{
   CODEC_LEN_t* pTmpCodLen;

   pTmpCodLen = (CODEC_LEN_t*)CODEC_LEN;

   /* look through table to get default packetization */
   do
   {
      if (nCodec == pTmpCodLen->nCodecName)
      {
         /* return found packetisation */
         return pTmpCodLen->nDefaultLenght;
      }
      pTmpCodLen++;
   }
   while (IFX_TAPI_ENC_TYPE_MAX != pTmpCodLen->nCodecName);

   return IFX_ERROR;
} /* VOIP_GetCodecDefaultFrameLen() */

/**
   Check if codec is available.

   \param pBoard     - polinter to board structure
   \param nCodType   - codec type number

   \return IFX_SUCCESS if codec supported, IFX_ERROR if not or error
*/
IFX_return_t VOIP_CodecAvailabilityCheck(BOARD_t* pBoard,
                                         IFX_TAPI_COD_TYPE_t nCodType)
{
   IFX_TAPI_CAP_t oCapList;
   IFX_int32_t nRet, nFd;
#ifdef TAPI_VERSION4
   IFX_int32_t i;
   BOARD_t* pTmpBoard;
#endif /* TAPI_VERSION4 */

   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PARAMETER_CHECK(((0>nCodType) || (IFX_TAPI_ENC_TYPE_MAX <= nCodType)),
                       nCodType, IFX_ERROR);
#ifdef TAPI_VERSION4
   /* for SVIP board resource manager is used, when XT16 is connected to SVIP
      coder channel from SVIP board will be used*/
   if (0 >= pBoard->nMaxCoderCh)
   {
      /* for all boards */
      for (i=0; i<pBoard->pCtrl->nBoardCnt; i++)
      {
         pTmpBoard = &pBoard->pCtrl->rgoBoards[i];
         /* data channels from this board will be used */
         if (0 < pBoard->nMaxCoderCh)
         {
            pBoard = pTmpBoard;
            break;
         } /* if (0 < pBoard->nMaxCoderCh) */
      } /* for all boards */
   } /* if (0 >= pBoard->nMaxCoderCh) */
#endif /* TAPI_VERSION4 */
   /* check if board has data channels */
   if (0 >= pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("%s: Err, VOIP_CodecAvailabilityCheck: "
             "board doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_ERROR;
   } /* if (0 >= pBoard->nMaxCoderCh) */
   /* asumption was made that first channel is a data channel */
   nFd = pBoard->nDevCtrl_FD;
   memset(&oCapList, 0, sizeof(IFX_TAPI_CAP_t));
   /* checking if codec available */
   oCapList.captype = IFX_TAPI_CAP_TYPE_CODEC;
   oCapList.cap = nCodType;

   /* check capability */
   TD_IOCTL(nRet, nFd, IFX_TAPI_CAP_CHECK, &oCapList,
            NO_PHONE_NUMBER, pBoard, 0, 0, TD_RETURN);
   /* if returned 0 then this capability is not supported */
   if (0 < nRet)
   {
      /* codec is supported */
      return IFX_SUCCESS;
   }
   TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
         ("%s: Err, Codec %s is not supported.\n",
          pBoard->pszBoardName, pCodecName[nCodType]));
   return IFX_ERROR;
} /* VOIP_CodecAvailabilityCheck() */

/**
   Starts the encoder/decoder.

   \param nDataCh  - data channel
   \param pConn - handle to connection structure
   \param pBoard - board handle

   \return IFX_SUCCESS on OK, otherwise IFX_ERROR
   \remarks for TAPI3 pConn is not used
            for TAPI4 nDataCh is not used
*/
IFX_return_t VOIP_StartCodec(IFX_int32_t nDataCh, CONNECTION_t* pConn,
                             BOARD_t* pBoard)
{
   IFX_int32_t data_fd = -1;
   IFX_TAPI_ENC_CFG_t oEncoderCfg;
   IFX_TAPI_DEC_CFG_t oDecoderCfg;
   IFX_return_t ret;
#ifdef TAPI_VERSION4
   IFX_TAPI_ENC_MODE_t encMode;
   IFX_TAPI_DEC_t dec;
#endif /* TAPI_VERSION4 */
#ifdef EASY336
   /* flag that is indicating that remote call is made
      between two phones on the same board */
   IFX_boolean_t fUsingBoardIp = IFX_FALSE;
   VOIP_DATA_CH_t *pOurCodec = IFX_NULL;
   VOIP_DATA_CH_t *pPeerCodec = IFX_NULL;
   SVIP_NAT_Rule_t NatRule;
   IFX_TAPI_PKT_NET_CFG_t PktNetCfg;
   /* used for printouts */
   struct in_addr nIpSrc, nIpDst;
#endif /* EASY336 */

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);

   if (0 >= pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_StartCodec: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_ERROR;
   }

#ifdef EASY336
   pOurCodec = Common_GetCodec(pBoard->pCtrl, pConn->voipID);
   if (pOurCodec == IFX_NULL)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get codec. "
             "(File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }
   if (pConn->nType == LOCAL_CALL)
   {
      pPeerCodec = Common_GetCodec(pBoard->pCtrl,
         pConn->oConnPeer.oLocal.pPhone->pVoipIDs[0]);
      if (pPeerCodec == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Phone No %d: Failed to get codec. "
                "(File: %s, line: %d)\n",
                pConn->oConnPeer.oLocal.pPhone->nPhoneNumber, 
                __FILE__, __LINE__));
         return IFX_ERROR;
      }

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Setting up a coder on %s device %d, channel %d to connect with\n"
          "coder on %s device %d, channel %d.\n",
         pBoard->pszBoardName,
         pOurCodec->nDev, pOurCodec->nCh,
         pBoard->pszBoardName,
         pPeerCodec->nDev, pPeerCodec->nCh));
   }
   else
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("Setting up a coder channel on %s device %d,\nchannel %d for VoIP "
             "session.\n",
             pBoard->pszBoardName,
             pOurCodec->nDev, pOurCodec->nCh));
#ifdef TAPI_VERSION4
      /* check if making external call to phone on the same board */
      if (pBoard->pCtrl->oTapidemo_IP_Addr.sin_addr.s_addr
          == pConn->oConnPeer.oRemote.oToAddr.sin_addr.s_addr)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Calling phone with the same IP address.\n"));
         fUsingBoardIp = IFX_TRUE;
      }
#endif
   }
#endif /* EASY336 */

#ifndef TAPI_VERSION4
   data_fd = VOIP_GetFD_OfCh(nDataCh, pBoard);
#else /* TAPI_VERSION4 */
   if (pBoard->fSingleFD)
   {
      data_fd = pBoard->nDevCtrl_FD;
   }
#endif /* TAPI_VERSION4 */

#ifdef TAPI_VERSION4
#ifdef EASY336
   if (pOurCodec->fStartCodec && !pOurCodec->fCodecStarted)
#endif /* EASY336 */
#else /* TAPI_VERSION4 */
   if (pBoard->pDataChStat[nDataCh].fStartCodec &&
       !pBoard->pDataChStat[nDataCh].fCodecStarted)
#endif /* TAPI_VERSION4 */
   {
#ifndef TAPI_VERSION4
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("DataCh No %d: Start encoder/decoder: %s,\n"
            "             packet length %s on dev %d\n",
            (int) nDataCh,
            pCodecName[(int)pBoard->pDataChStat[nDataCh].nCodecType],
            oFrameLen[(int)pBoard->pDataChStat[nDataCh].nFrameLen],
            (int) data_fd));
#else /* TAPI_VERSION4 */
#ifdef EASY336
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("Start encoder/decoder on dev %d, data ch %d\n",
            (int) pOurCodec->nDev, (int) pOurCodec->nCh));
#endif /* EASY336 */
#endif /* TAPI_VERSION4 */

#ifndef STREAM_1_1

      /* From the version 1.2.7.0 of TAPI drivers both codec and frame
        length should be changed by the IFX_TAPI_ENC_CFG_SET function */
#ifndef TAPI_VERSION4
      oEncoderCfg.nEncType = pBoard->pDataChStat[nDataCh].nCodecType;
      oEncoderCfg.nFrameLen = pBoard->pDataChStat[nDataCh].nFrameLen;
      oEncoderCfg.AAL2BitPack = pBoard->pDataChStat[nDataCh].nBitPack;
#else /*TAPI_VERSION4*/
#ifdef EASY336
      oEncoderCfg.nEncType = pOurCodec->nCodecType;
      oEncoderCfg.nFrameLen = pOurCodec->nFrameLen;
      oEncoderCfg.AAL2BitPack = pOurCodec->nBitPack;
#endif /* EASY336 */
#endif /*TAPI_VERSION4*/
      if (oEncoderCfg.AAL2BitPack == IFX_TAPI_COD_AAL2_BITPACK)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("Using ITU-T I366.2 Bit Alignment for G.726 codecs\n"));
      }

#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
#ifdef EASY336
         oEncoderCfg.dev = pOurCodec->nDev;
         oEncoderCfg.ch = pOurCodec->nCh;
#endif /* EASY336 */
      }
#endif /*TAPI_VERSION4*/

      /* Set coder type and packetisation time */
#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
         TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_ENC_CFG_SET,
            (IFX_int32_t) &oEncoderCfg)) == IFX_SUCCESS, data_fd,
            pBoard, oEncoderCfg.dev, oEncoderCfg.ch);
      } /* if (pBoard->fSingleFD) */
      else
#endif /*TAPI_VERSION4*/
      {
         ret = ioctl(data_fd, IFX_TAPI_ENC_CFG_SET, (int) &oEncoderCfg);
      } /* if (pBoard->fSingleFD) */
      if (ret != IFX_SUCCESS)
      {
#ifndef TAPI_VERSION4
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Setting encoder type %s with frame length "
               "(packetisation time) %s failed. (File: %s, line: %d)\n",
               pCodecName[(int)pBoard->pDataChStat[nDataCh].nCodecType],
               oFrameLen[(int)pBoard->pDataChStat[nDataCh].nFrameLen],
               __FILE__, __LINE__));
#else /*TAPI_VERSION4*/
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Setting encoder type %d with frame length id "
               "(packetisation time) %d failed. (File: %s, line: %d)\n",
               oEncoderCfg.nEncType, oEncoderCfg.nFrameLen,
               __FILE__, __LINE__));
#endif /*TAPI_VERSION4*/
         return IFX_ERROR;
      }
#endif /* STREAM_1_1 */
      /* set decoder configuration, basicaly needed only by
         ITU-T I366.2 Bit Alignment for G.726 */
      oDecoderCfg.AAL2BitPack = oEncoderCfg.AAL2BitPack;
      /* packet loss handled according to codec type - only TAPI V3 */
      oDecoderCfg.Plc = IFX_TAPI_DEC_PLC_CODEC;
#ifdef TAPI_VERSION4
      oDecoderCfg.ch = oEncoderCfg.ch;
      oDecoderCfg.dev = oEncoderCfg.dev;
#endif /*TAPI_VERSION4*/
      /* Set decoder type */
#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
         TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_DEC_CFG_SET,
            (IFX_int32_t) &oDecoderCfg)) == IFX_SUCCESS, data_fd,
            pBoard, oDecoderCfg.dev, oDecoderCfg.ch);
      } /* if (pBoard->fSingleFD) */
      else
#endif /*TAPI_VERSION4*/
      {
         ret = ioctl(data_fd, IFX_TAPI_DEC_CFG_SET, (int) &oDecoderCfg);
      } /* if (pBoard->fSingleFD) */
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err, Failed to configure decoder. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }

#ifdef EASY336 /* place Routing setup here */
      if (pConn->nType == LOCAL_CALL)
      {
         if (SVIP_libVoIP_CFG_Set(pOurCodec->nDev, pOurCodec->nCh, IFX_TRUE,
            htons(pOurCodec->nUDP_Port), htons(pPeerCodec->nUDP_Port), 
            pPeerCodec->nDev, fUsingBoardIp, IFX_NULL, 0, &PktNetCfg,
            &NatRule) !=
            SVIP_libStatusOk)
            return IFX_ERROR;
      }
      else
      {
         if (SVIP_libVoIP_CFG_Set(pOurCodec->nDev, pOurCodec->nCh, 
            IFX_FALSE,
            htons(pConn->oUsedAddr.sin_port), 
            htons(pConn->oConnPeer.oRemote.oToAddr.sin_port), 
            0, fUsingBoardIp, pConn->oConnPeer.oRemote.MAC, 
            pConn->oConnPeer.oRemote.oToAddr.sin_addr.s_addr, &PktNetCfg,
            &NatRule) !=
            SVIP_libStatusOk)
            return IFX_ERROR;
      }

      memset(&nIpDst, 0, sizeof(nIpDst));
      memset(&nIpSrc, 0, sizeof(nIpSrc));

      /* copy data for easier printout */
      nIpSrc.s_addr = PktNetCfg.nIp.v4.IpSrc;
      nIpDst.s_addr = PktNetCfg.nIp.v4.IpDst;

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("####### Configure Ethernet session ########\n"
             "\tDestination MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
             "\tSource      MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
             "\tPorts: Rec-%d, Des-%d, Sou-%d\n"
             "\tchan %d, dev %d\n"
             "\tType: %s, Using: %s\n"
             "\tIP Dst %s,\n",
             PktNetCfg.MacDst[0],
             PktNetCfg.MacDst[1],
             PktNetCfg.MacDst[2],
             PktNetCfg.MacDst[3],
             PktNetCfg.MacDst[4],
             PktNetCfg.MacDst[5],
             PktNetCfg.MacSrc[0],
             PktNetCfg.MacSrc[1],
             PktNetCfg.MacSrc[2],
             PktNetCfg.MacSrc[3],
             PktNetCfg.MacSrc[4],
             PktNetCfg.MacSrc[5],
             PktNetCfg.nUdpRcv,
             PktNetCfg.nUdpDst,
             PktNetCfg.nUdpSrc,
             PktNetCfg.ch, PktNetCfg.dev,
             PktNetCfg.nPktType == Voice ? "Voice" :
             PktNetCfg.nPktType == Fax ? "Fax" : "FaxTrace",
             PktNetCfg.nIpMode == IFX_TAPI_PKT_NET_IPV4 ? "IPV4" : "IPV6",
             inet_ntoa(nIpDst)));
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("\tIP Src %s\n"
             "\tnVLAN_Tagging %d\n"
             "###############\n",
             inet_ntoa(nIpSrc),
             PktNetCfg.nVLAN_Tagging));

      TD_ASSERT ((ret = ioctl (data_fd, IFX_TAPI_PKT_NET_CFG_SET,
         (IFX_int32_t)&PktNetCfg)) == IFX_SUCCESS, data_fd, pBoard,
         PktNetCfg.dev, PktNetCfg.ch);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Session packet header information configuration for device "
                "%d, channel %d failed. "
                "(File: %s, line: %d)\n",
                PktNetCfg.dev, PktNetCfg.ch,
                __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset(&nIpDst, 0, sizeof(nIpDst));
      memset(&nIpSrc, 0, sizeof(nIpSrc));

      nIpDst.s_addr = NatRule.remIP;
      nIpSrc.s_addr = NatRule.locIP;

      /* SVIP library is signalling that no NAT rule should be added by setting
       * port value to 0 */
      if (NatRule.locUDP != 0)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("####### Adding NAT rule ########\n"
                "\tDestination MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
                "\tSource      MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
                "\tUDP port: %d\n"
                "\tIP Dst %s,\n",
                NatRule.remMAC[0],
                NatRule.remMAC[1],
                NatRule.remMAC[2],
                NatRule.remMAC[3],
                NatRule.remMAC[4],
                NatRule.remMAC[5],
                NatRule.locMAC[0],
                NatRule.locMAC[1],
                NatRule.locMAC[2],
                NatRule.locMAC[3],
                NatRule.locMAC[4],
                NatRule.locMAC[5],
                NatRule.locUDP,
                inet_ntoa(nIpDst)));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("\tIP Src %s\n"
                "################################\n",
                inet_ntoa(nIpSrc)));
   
         ret = SVIP_NAT_RuleAdd (&NatRule);
         if (ret != SVIP_libStatusOk)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Adding SVIP NAT rule for device %d, channel %d failed. "
                  "(File: %s, line: %d)\n",
                  pOurCodec->nDev,
                  pOurCodec->nCh,
                  __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }
#endif /*EASY336*/

      /* Start encoding */
#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
#ifdef EASY336
         encMode.dev = pOurCodec->nDev;
         encMode.ch = pOurCodec->nCh;
#endif /* EASY336 */
         TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_ENC_START,
            (IFX_int32_t)&encMode)) == IFX_SUCCESS, data_fd, pBoard,
            encMode.dev, encMode.ch);
      } /* if (pBoard->fSingleFD) */
      else
#endif /* TAPI_VERSION4 */
      {
         ret = ioctl(data_fd, IFX_TAPI_ENC_START, 0);
      } /* if (pBoard->fSingleFD) */
      if ( IFX_ERROR == ret )
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Start encoder failed. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }

      /* Start decoding */
#ifdef TAPI_VERSION4
      if (pBoard->fSingleFD)
      {
#ifdef EASY336
         dec.dev = pOurCodec->nDev;
         dec.ch = pOurCodec->nCh;
#endif /* EASY336 */
         TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_DEC_START,
            (IFX_int32_t) &dec)) == IFX_SUCCESS, data_fd, pBoard,
            dec.dev, dec.ch);
      } /* if (pBoard->fSingleFD) */
      else
#endif /* TAPI_VERSION4 */
      {
         ret = ioctl(data_fd, IFX_TAPI_DEC_START, 0);
      } /* if (pBoard->fSingleFD) */
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Start decoder failed. (File: %s, line: %d)\n",
               __FILE__, __LINE__));
         return IFX_ERROR;
      }

#ifndef TAPI_VERSION4
      pBoard->pDataChStat[nDataCh].fCodecStarted = IFX_TRUE;
#else /* TAPI_VERSION4 */
#ifdef EASY336
      pOurCodec->fCodecStarted = IFX_TRUE;
#endif
#endif /* TAPI_VERSION4 */
   } /* if */

   return IFX_SUCCESS;
} /* VOIP_StartCodec() */

/**
   Stops the encoder/decoder.

   \param nDataCh  - data channel
   \param pConn - handle to connection structure
   \param pBoard - board handle

   \return IFX_SUCCESS on ok, otherwise IFX_ERROR
   \remarks for TAPI3 pConn is not used
            for TAPI4 nDataCh is not used
*/
IFX_return_t VOIP_StopCodec(IFX_int32_t nDataCh, CONNECTION_t* pConn,
                            BOARD_t* pBoard)
{
   IFX_int32_t data_fd = -1;
   IFX_return_t ret;
#ifdef TAPI_VERSION4
   VOIP_DATA_CH_t *pCodec = IFX_NULL;
   IFX_TAPI_ENC_MODE_t encMode;
   IFX_TAPI_DEC_t dec;
#ifdef EASY336
   /* flag that is indicating that the call is made
      between two phones on the same board */
   IFX_boolean_t fUsingBoardIp = IFX_FALSE;
#endif /* EASY336 */
#endif /* TAPI_VERSION4 */
#ifdef EASY336
   SVIP_NAT_Rule_t NatRule;
#endif /* EASY336 */
   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);

   if (0 >= pBoard->nMaxCoderCh)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_StopCodec: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_ERROR;
   }

#ifdef EASY336
   memset(&NatRule, 0, sizeof(SVIP_NAT_Rule_t));
   /* check if making external call to phone on the same board */
   if (pBoard->pCtrl->oTapidemo_IP_Addr.sin_addr.s_addr
       == pConn->oConnPeer.oRemote.oToAddr.sin_addr.s_addr)
   {
      /* Call between phones with the same IP address. */
      fUsingBoardIp = IFX_TRUE;
   }
#endif /* EASY336 */

#ifndef TAPI_VERSION4
   data_fd = VOIP_GetFD_OfCh(nDataCh, pBoard);
#else/* TAPI_VERSION4 */
   if (pBoard->fSingleFD)
      data_fd = pBoard->nDevCtrl_FD;
#endif/* TAPI_VERSION4 */

#ifdef EASY336
   pCodec = Common_GetCodec(pBoard->pCtrl, pConn->voipID);
   if (pCodec == IFX_NULL)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get codec. "
             "(File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 */

#ifndef TAPI_VERSION4
   if (IFX_TRUE == pBoard->pDataChStat[nDataCh].fCodecStarted)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("DataCh No %d: Stop encoder/decoder on fd %d\n",
           (int) nDataCh, (int) data_fd));
   }
#else /* TAPI_VERSION4 */
   if (IFX_TRUE == pCodec->fCodecStarted)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
           ("DataCh No %d: Stop encoder/decoder on dev %d\n",
           (int) pCodec->nCh, (int) pCodec->nDev));
   }
#endif /* TAPI_VERSION4 */

      /* Stop encoding */
#ifdef TAPI_VERSION4
   if (pBoard->fSingleFD)
   {
      encMode.dev = pCodec->nDev;
      encMode.ch = pCodec->nCh;
      TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_ENC_STOP,
         (IFX_int32_t)&encMode)) == IFX_SUCCESS, data_fd, pBoard,
         encMode.dev, encMode.ch);
   } /* if (pBoard->fSingleFD) */
   else
#endif /* TAPI_VERSION4 */
   {
      ret = ioctl(data_fd, IFX_TAPI_ENC_STOP, 0);
   } /* if (pBoard->fSingleFD) */
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Stop encoder failed. (File: %s, line: %d)\n",
           __FILE__, __LINE__));
      return IFX_ERROR;
   }

   /* Stop decoding */
#ifdef TAPI_VERSION4
   if (pBoard->fSingleFD)
   {
      dec.dev = pCodec->nDev;
      dec.ch = pCodec->nCh;
      TD_ASSERT ((ret = ioctl(data_fd, IFX_TAPI_DEC_STOP,
         (IFX_uint32_t)&dec)) == IFX_SUCCESS, data_fd, pBoard,
         dec.dev, dec.ch);
   } /* if (pBoard->fSingleFD) */
   else
#endif/* TAPI_VERSION4 */
   {
      ret = ioctl(data_fd, IFX_TAPI_DEC_STOP, 0);
   } /* if (pBoard->fSingleFD) */
   if (IFX_ERROR == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Stop decoder failed. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }

#ifndef TAPI_VERSION4
   pBoard->pDataChStat[nDataCh].fCodecStarted = IFX_FALSE;
#else/* TAPI_VERSION4 */
   pCodec->fCodecStarted = IFX_FALSE;
#endif/* TAPI_VERSION4 */

#ifdef EASY336

   if (pConn->nType == LOCAL_CALL)
      NatRule.locUDP = htons(pCodec->nUDP_Port);
   else
      NatRule.locUDP = pConn->oUsedAddr.sin_port;

   if (SVIP_NAT_RuleRemove (&NatRule) != SVIP_libStatusOk)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Removing SVIP NAT rule for device %d, channel %d failed. "
            "(File: %s, line: %d)\n",
            pCodec->nDev,
            pCodec->nCh,
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 */

   return IFX_SUCCESS;
} /* VOIP_StopCodec() */

#ifdef TAPI_VERSION4
/**
   Starts the fax channel.

   \param pCtrl - handle to control structure
   \param pConn  - pointer to phone connection
   \param pBoard - board handle

   \return IFX_SUCCESS on OK, otherwise IFX_ERROR
*/
IFX_return_t VOIP_FaxStart(CTRL_STATUS_t* pCtrl, CONNECTION_t* pConn,
                           BOARD_t* pBoard)
{
#ifdef EASY336
   IFX_return_t ret;
   SVIP_NAT_Rule_t NatRule;
   IFX_TAPI_PKT_NET_CFG_t PktNetCfg;
   VOIP_DATA_CH_t *pOurCodec = IFX_NULL;
   VOIP_DATA_CH_t *pPeerCodec = IFX_NULL;
   /** \todo: add this strucutre to TAPI3 */
   IFX_TAPI_T38_SESS_CFG_t t38SessCfg;
   /* used for printouts */
   struct in_addr nIpSrc, nIpDst;
   /* flag that is indicating that remote call is made
      between two phones on the same board */
   IFX_boolean_t fUsingBoardIp = IFX_FALSE;
   
   memset (&t38SessCfg, 0, sizeof (IFX_TAPI_T38_SESS_CFG_t));
#endif
   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);
   TD_PTR_CHECK(pCtrl, IFX_ERROR);

   if (pBoard->nMaxCoderCh == 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_FaxStart: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

#ifdef EASY336
   pOurCodec = Common_GetCodec(pCtrl, pConn->voipID);
   if (pOurCodec == IFX_NULL)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Failed to get codec. "
             "(File: %s, line: %d)\n",
             __FILE__, __LINE__));
      return IFX_ERROR;
   }

   if (pConn->nType == LOCAL_CALL)
   {
      pPeerCodec = Common_GetCodec(pCtrl,
         pConn->oConnPeer.oLocal.pPhone->pVoipIDs[0]);
      if (pPeerCodec == IFX_NULL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Failed to get codec. "
                "(File: %s, line: %d)\n",
                __FILE__, __LINE__));
         return IFX_ERROR;
      }
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Setting up a FAX on device %d, channel %d to connect with FAX\n"
         "on device %d, channel %d.\n",
         pOurCodec->nDev, pOurCodec->nCh,
         pPeerCodec->nDev, pPeerCodec->nCh));
   }
   else
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
         ("Setting up a FAX channel on device %d, channel %d\nfor VoIP "
         "session.\n",
         pOurCodec->nDev, pOurCodec->nCh));

   if (pOurCodec->fStartCodec && !pOurCodec->fCodecStarted)
   {
      if (pConn->nType == LOCAL_CALL)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW, (
               "Local FAX connection\n"));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));
      }
      else
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW, (
                     "VoIP FAX connection\n"));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));
      }

      if (pConn->nType == LOCAL_CALL)
      {
         if (SVIP_libVoIP_CFG_Set(pOurCodec->nDev, pOurCodec->nCh, IFX_TRUE,
            htons(pOurCodec->nUDP_Port), htons(pPeerCodec->nUDP_Port), 
            pPeerCodec->nDev, fUsingBoardIp, IFX_NULL, 0, &PktNetCfg,
            &NatRule) !=
            SVIP_libStatusOk)
            return IFX_ERROR;
      }
      else
      {
         if (SVIP_libVoIP_CFG_Set(pOurCodec->nDev, pOurCodec->nCh, 
            IFX_FALSE,
            htons(pConn->oUsedAddr.sin_port), 
            htons(pConn->oConnPeer.oRemote.oToAddr.sin_port), 
            0, fUsingBoardIp, pConn->oConnPeer.oRemote.MAC, 
            pConn->oConnPeer.oRemote.oToAddr.sin_addr.s_addr, &PktNetCfg,
            &NatRule) !=
            SVIP_libStatusOk)
            return IFX_ERROR;
      }
      PktNetCfg.nPktType = Fax;

      memset(&nIpDst, 0, sizeof(nIpDst));
      memset(&nIpSrc, 0, sizeof(nIpSrc));

      /* copy data for easier printout */
      nIpSrc.s_addr = PktNetCfg.nIp.v4.IpSrc;
      nIpDst.s_addr = PktNetCfg.nIp.v4.IpDst;

      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("####### Configure Ethernet session ########\n"
             "\tDestination MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
             "\tSource      MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
             "\tPorts: Rec-%d, Des-%d, Sou-%d\n"
             "\tchan %d, dev %d\n"
             "\tType: %s, Using: %s\n"
             "\tIP Dst %s,\n",
             PktNetCfg.MacDst[0],
             PktNetCfg.MacDst[1],
             PktNetCfg.MacDst[2],
             PktNetCfg.MacDst[3],
             PktNetCfg.MacDst[4],
             PktNetCfg.MacDst[5],
             PktNetCfg.MacSrc[0],
             PktNetCfg.MacSrc[1],
             PktNetCfg.MacSrc[2],
             PktNetCfg.MacSrc[3],
             PktNetCfg.MacSrc[4],
             PktNetCfg.MacSrc[5],
             PktNetCfg.nUdpRcv,
             PktNetCfg.nUdpDst,
             PktNetCfg.nUdpSrc,
             PktNetCfg.ch, PktNetCfg.dev,
             PktNetCfg.nPktType == Voice ? "Voice" :
             PktNetCfg.nPktType == Fax ? "Fax" : "FaxTrace",
             PktNetCfg.nIpMode == IFX_TAPI_PKT_NET_IPV4 ? "IPV4" : "IPV6",
             inet_ntoa(nIpDst)));
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("\tIP Src %s\n"
             "\tnVLAN_Tagging %d\n"
             "###############\n",
             inet_ntoa(nIpSrc),
             PktNetCfg.nVLAN_Tagging));
      
      TD_ASSERT ((ret = ioctl (pBoard->nDevCtrl_FD, IFX_TAPI_PKT_NET_CFG_SET,
         (IFX_int32_t)&PktNetCfg)) == IFX_SUCCESS, pBoard->nDevCtrl_FD, pBoard,
         PktNetCfg.dev, PktNetCfg.ch);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
               ("Session packet header information configuration for device "
                "%d, channel %d failed. "
                "(File: %s, line: %d)\n",
                PktNetCfg.dev, PktNetCfg.ch,
                __FILE__, __LINE__));
         return IFX_ERROR;
      }

      memset(&nIpDst, 0, sizeof(nIpDst));
      memset(&nIpSrc, 0, sizeof(nIpSrc));

      nIpDst.s_addr = NatRule.remIP;
      nIpSrc.s_addr = NatRule.locIP;

      /* SVIP library is signalling that no NAT rule should be added by setting
       * port value to 0 */
      if (NatRule.locUDP != 0)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("####### Adding NAT rule ########\n"
                "\tDestination MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
                "\tSource      MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
                "\tUDP port: %d\n"
                "\tIP Dst %s,\n",
                NatRule.remMAC[0],
                NatRule.remMAC[1],
                NatRule.remMAC[2],
                NatRule.remMAC[3],
                NatRule.remMAC[4],
                NatRule.remMAC[5],
                NatRule.locMAC[0],
                NatRule.locMAC[1],
                NatRule.locMAC[2],
                NatRule.locMAC[3],
                NatRule.locMAC[4],
                NatRule.locMAC[5],
                NatRule.locUDP,
                inet_ntoa(nIpDst)));
         TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("\tIP Src %s\n"
                "################################\n",
                inet_ntoa(nIpSrc)));
   
         ret = SVIP_NAT_RuleAdd (&NatRule);
         if (ret != SVIP_libStatusOk)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Adding SVIP NAT rule for device %d, channel %d failed. "
                  "(File: %s, line: %d)\n",
                  pOurCodec->nDev,
                  pOurCodec->nCh,
                  __FILE__, __LINE__));
            return IFX_ERROR;
         }
      }
      
      t38SessCfg.dev = pOurCodec->nDev;
      t38SessCfg.ch = pOurCodec->nCh;
      t38SessCfg.nRateManagement = IFX_TAPI_T38_TRANS_TCF;
      t38SessCfg.nProtocol = IFX_TAPI_T38_UDP;
      t38SessCfg.nBitRateMax = 14400;
      t38SessCfg.nUDPErrCorr = IFX_TAPI_T38_RED;
      t38SessCfg.nUDPBuffSizeMax = 1000;
      t38SessCfg.nUDPDatagramSizeMax = 1000;
      t38SessCfg.nT38Ver = 0;
      /* Start session with predefined T.38 capabilities t38SessCfg */
      if (pBoard->fSingleFD)
      {
         TD_ASSERT ((ret = ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_T38_SESS_START,
            (IFX_int32_t) &t38SessCfg)) == IFX_SUCCESS, pBoard->nDevCtrl_FD,
            pBoard, t38SessCfg.dev, t38SessCfg.ch);
      } /* if (pBoard->fSingleFD) */
      if (ret != IFX_SUCCESS)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("FAX Start failed. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
         return IFX_ERROR;
      }
      pOurCodec->fCodecStarted = IFX_TRUE;
   } /* if */
#endif /* EASY336 */
   return IFX_SUCCESS;
} /* VOIP_FaxStart() */

/**
   Stopss the fax channel.

   \param pCtrl - handle to control structure
   \param pConn  - pointer to phone connection
   \param pBoard - board handle

   \return IFX_SUCCESS on OK, otherwise IFX_ERROR
*/
IFX_return_t VOIP_FaxStop(CTRL_STATUS_t* pCtrl, CONNECTION_t* pConn,
                          BOARD_t* pBoard)
{
#ifdef EASY336
   SVIP_NAT_Rule_t NatRule;
   VOIP_DATA_CH_t *pCodec = IFX_NULL;
   /** \todo: add these strucutre to TAPI3 */
   IFX_TAPI_T38_SESS_STOP_t t38SessStop;
   IFX_TAPI_T38_SESS_STATISTICS_t t38SessStat;
   IFX_return_t ret = IFX_ERROR;

   memset (&t38SessStat, 0, sizeof (IFX_TAPI_T38_SESS_STATISTICS_t));
   memset (&t38SessStop, 0, sizeof (IFX_TAPI_T38_SESS_STOP_t));
#endif /* EASY336 */
   
   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);
   TD_PTR_CHECK(pCtrl, IFX_ERROR);

   if (pBoard->nMaxCoderCh == 0)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_FaxStop: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

#ifdef EASY336
   memset(&NatRule, 0, sizeof(SVIP_NAT_Rule_t));
   pCodec = Common_GetCodec(pCtrl, pConn->voipID);
   if (pCodec == IFX_NULL)
      return IFX_ERROR;

   /* Stop fax */
   t38SessStat.dev = pCodec->nDev;
   t38SessStat.ch = pCodec->nCh;
   if (pBoard->fSingleFD)
   {
      TD_ASSERT ((ret = ioctl(pBoard->nDevCtrl_FD, 
         IFX_TAPI_T38_SESS_STATISTICS_GET,
         (IFX_int32_t) &t38SessStat)) == IFX_SUCCESS, pBoard->nDevCtrl_FD,
         pBoard, t38SessStat.dev, t38SessStat.ch);
   } /* if (pBoard->fSingleFD) */
   if (ret != IFX_SUCCESS)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("T38 get stat failed. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));
   TRACE(TAPIDEMO, DBG_LEVEL_LOW, (
         "Fax session channel %d statistics:\n"
         "nFaxSessState %d\n"
         "nT38VerMajor %d\n"
         "nT38VerMin %d\n"
         "SessInfo %d\n"
         "nFdpStand %d\n"
         "nPktLost %d\n"
         "nPktRecov %d\n"
         "nPktGroupLost %d\n"
         "nFttRsp %d\n"
         "nPagesTx %d\n"
         "nLineBreak %d\n"
         "nV21FrmBreak %d\n"
         "nEcmFrmBreak %d\n",
         pCodec->nCh,
         t38SessStat.nFaxSessState,
         t38SessStat.nT38VerMajor,
         t38SessStat.nT38VerMin,
         t38SessStat.SessInfo,
         t38SessStat.nFdpStand,
         t38SessStat.nPktLost,
         t38SessStat.nPktRecov,
         t38SessStat.nPktGroupLost,
         t38SessStat.nFttRsp,
         t38SessStat.nPagesTx,
         t38SessStat.nLineBreak,
         t38SessStat.nV21FrmBreak,
         t38SessStat.nEcmFrmBreak
         ));
   TRACE(TAPIDEMO, DBG_LEVEL_LOW,
               ("############################################\n"));

   t38SessStop.dev = pCodec->nDev;
   t38SessStop.ch = pCodec->nCh;
   if (IFX_ERROR == ioctl(pBoard->nDevCtrl_FD, IFX_TAPI_T38_SESS_STOP,
      (IFX_int32_t)&t38SessStop))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("T38 Stop failed. (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   pCodec->fCodecStarted = IFX_FALSE;
   
   if (pConn->nType == LOCAL_CALL)
      NatRule.locUDP = htons(pCodec->nUDP_Port);
   else
      NatRule.locUDP = pConn->oUsedAddr.sin_port;

   if (SVIP_NAT_RuleRemove (&NatRule) != SVIP_libStatusOk)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Removing SVIP NAT rule for device %d, channel %d failed. "
            "(File: %s, line: %d)\n",
            pCodec->nDev,
            pCodec->nCh,
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
#endif /* EASY336 */
   return IFX_SUCCESS;
} /* VOIP_FaxStop() */
#endif /* TAPI_VERSION4 */

/**
   Set codec flag if coder can be started or not.

   \param nDataCh - data channel
   \param fStartCodec - IFX_TRUE set flag to IFX_TRUE,
                        or IFX_FALSE is IFX_FALSE
   \param pBoard - board handle

   \return Flag is set to proper value
*/
IFX_void_t VOIP_SetCodecFlag(IFX_int32_t nDataCh, IFX_boolean_t fStartCodec,
                             BOARD_t* pBoard)
{

   if ((IFX_NULL != pBoard) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_SetCodecFlag: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return;
   }

   if ((pBoard == IFX_NULL) || (0 > nDataCh) ||
       (pBoard->nMaxCoderCh <= nDataCh) || (pBoard->pDataChStat == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }

   pBoard->pDataChStat[nDataCh].fStartCodec = fStartCodec;
} /* VOIP_SetCodecFlag() */


/**
   Return free data channel.

   \param pBoard - pointer to board

   \return free data channel or NO_FREE_DATA_CH
*/
IFX_int32_t VOIP_GetFreeDataCh(BOARD_t* pBoard)
{
   IFX_uint32_t i = 0;
   IFX_int32_t free_ch = NO_FREE_DATA_CH;

   if ((IFX_NULL != pBoard) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_GetFreeDataCh: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);

   for (i = 0; i < pBoard->nMaxCoderCh; i++)
   {
      if (IFX_FALSE == pBoard->pDataChStat[i].fMapped)
      {
         free_ch = i;
         break;
      }
   }

   return free_ch;
} /* VOIP_GetFreeDataCh() */


/**
   Free mapped data channel.

   \param nDataCh - data channel number
   \param pBoard - board handle

   \return IFX_SUCCESS - data channel is freed, otherwise IFX_ERROR
*/
IFX_return_t VOIP_FreeDataCh(IFX_int32_t nDataCh, BOARD_t* pBoard)
{

   if ((IFX_NULL != pBoard) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_VOIP_FreeDataCh: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nDataCh), nDataCh, IFX_ERROR);
   TD_PARAMETER_CHECK((pBoard->nMaxCoderCh <= nDataCh), nDataCh, IFX_ERROR);

   pBoard->pDataChStat[nDataCh].fMapped = IFX_FALSE;

   return IFX_SUCCESS;
} /* VOIP_FreeDataCh() */


/**
   Reserve data channel as mapped.

   \param nDataCh - data channel number
   \param pBoard - pointer to board

   \return IFX_SUCCESS - data channel is reserved, otherwise IFX_ERROR
*/
IFX_return_t VOIP_ReserveDataCh(IFX_int32_t nDataCh, BOARD_t* pBoard)
{
   if ((pBoard != IFX_NULL) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_ReserveDataCh: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nDataCh), nDataCh, IFX_ERROR);
   TD_PARAMETER_CHECK((pBoard->nMaxCoderCh <= nDataCh), nDataCh, IFX_ERROR);

   if (IFX_TRUE == pBoard->pDataChStat[nDataCh].fMapped)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Data ch %d, already reserved\n", (int) nDataCh));
      return IFX_ERROR;
   }

   pBoard->pDataChStat[nDataCh].fMapped = IFX_TRUE;

   return IFX_SUCCESS;
} /* VOIP_GetFreeDataCh() */


/**
   Return free VOIP socket.

   \param pCtrl - handle to control status structure

   \return socket or NO_SOCKET on error
*/
IFX_int32_t VOIP_GetFreeSocket(CTRL_STATUS_t* pCtrl)
{
   IFX_uint32_t i = 0;
   IFX_int32_t free_socket = NO_SOCKET;

   /* check input parameters */
   TD_PTR_CHECK(pCtrl, NO_SOCKET);
   TD_PTR_CHECK(pCtrl->rgnSockets, NO_SOCKET);

   for (i = 0; i < pCtrl->nSumCoderCh; i++)
   {
      if (IFX_FALSE == pCtrl->rgnSockets[i].fUsed)
      {
         free_socket = pCtrl->rgnSockets[i].nSocket;
         pCtrl->rgnSockets[i].fUsed = IFX_TRUE;
         break;
      }
   }

   return free_socket;
} /* VOIP_GetFreeSocket() */


/**
   Release VOIP socket.

   \param pCtrl - handle to control status structure

  \return socket or NO_SOCKET on error
*/
IFX_void_t VOIP_ReleaseSocket(CTRL_STATUS_t* pCtrl)
{
   IFX_uint32_t i = 0;


   if ((IFX_NULL == pCtrl) || (pCtrl->rgnSockets == IFX_NULL))
   {
      /* Wrong input arguments */
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("Invalid input argument(s). (File: %s, line: %d)\n",
            __FILE__, __LINE__));
      return;
   }

   for (i = 0; i < pCtrl->nSumCoderCh; i++)
   {
      if (pCtrl->rgnSockets[i].nSocket < 0)
      {
         /* This one is not open, skip it. */
         TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Socket not open, skip it.\n"));
         continue;
      }

      if (close(pCtrl->rgnSockets[i].nSocket) < 0)
      {
         /* Err closing socket */
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Err closing socket with fd %d. (File: %s, line: %d)\n",
               (int) pCtrl->rgnSockets[i].nSocket, __FILE__, __LINE__));
      }
   }
} /* VOIP_ReleaseSocket() */


/**
   Return VOIP socket according to channel idx.

   \param nChIdx - channel index
   \param pBoard - board handle

   \return socket or NO_SOCKET on error
*/
IFX_int32_t VOIP_GetSocket(IFX_int32_t nChIdx, BOARD_t* pBoard)
{
   if ((pBoard != IFX_NULL) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_GetSocket: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

   /* check input parameters */
   TD_PTR_CHECK(pBoard, NO_SOCKET);
   TD_PTR_CHECK(pBoard->rgnSockets, NO_SOCKET);
   TD_PARAMETER_CHECK((0 > nChIdx), nChIdx, NO_SOCKET);
   TD_PARAMETER_CHECK((pBoard->nMaxCoderCh <= nChIdx), nChIdx, NO_SOCKET);

   return pBoard->rgnSockets[nChIdx];
} /* VOIP_GetSocket() */


/**
   Return VOIP port which is used by this socket.

   \param nSocket - socket number
   \param pCtrl - pointer to control structure

   \return socket or NO_SOCKET on error
*/
IFX_int32_t VOIP_GetSocketPort(IFX_int32_t nSocket, CTRL_STATUS_t* pCtrl)
{
   IFX_int32_t i = 0;

   /* check input parameters */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pCtrl->rgnSockets, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nSocket), nSocket, IFX_ERROR);

   for (i = 0; i < pCtrl->nSumCoderCh; i++)
   {
      if (pCtrl->rgnSockets[i].nSocket == nSocket)
      {
         return pCtrl->rgnSockets[i].nPort;
      }
   }

   return -1;
} /* VOIP_GetSocket() */


/**
   Map phone channel to data channel.

   \param nDataCh     - target data channel
   \param nDev        - device number
   \param nAddPhoneCh - which phone channel to add
   \param fDoMapping - IFX_TRUE do mapping, IFX_FALSE do unmapping
   \param pBoard - board handle

   \return IFX_SUCCESS if ok, otherwise IFX_ERROR
   \remarks nDev is used only for TAPI_VERSION4
*/
IFX_return_t VOIP_MapPhoneToData(IFX_int32_t nDataCh,
                                 IFX_uint16_t nDev,
                                 IFX_int32_t nAddPhoneCh,
                                 IFX_boolean_t fDoMapping,
                                 BOARD_t* pBoard)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_TAPI_MAP_DATA_t datamap;
   IFX_int32_t fd_ch = -1;


   if ((pBoard != IFX_NULL) && (0 == pBoard->nMaxCoderCh))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_LOW,
            ("VOIP_MapPhoneToData: board %s doesn't have coder channels.\n",
             pBoard->pszBoardName));
      return IFX_SUCCESS;
   }

   /* check input parameters */
   TD_PTR_CHECK(pBoard, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nDataCh), nDataCh, IFX_ERROR);
   TD_PARAMETER_CHECK((pBoard->nMaxCoderCh <= nDataCh), nDataCh, IFX_ERROR);
   TD_PARAMETER_CHECK((0 > nAddPhoneCh), nAddPhoneCh, IFX_ERROR);
   TD_PARAMETER_CHECK((pBoard->nMaxAnalogCh < nAddPhoneCh), nAddPhoneCh, 
      IFX_ERROR);

   memset(&datamap, 0, sizeof(IFX_TAPI_MAP_DATA_t));

#ifdef TAPI_VERSION4
   if (!pBoard->fSingleFD)
#endif /* TAPI_VERSION4 */
   {
      /* Get file descriptor for this channel */
      fd_ch = VOIP_GetFD_OfCh(nDataCh, pBoard);
   }

#ifdef TAPI_VERSION4
   datamap.dev = nDev;
   datamap.ch = nDataCh;
#endif /* TAPI_VERSION4 */
   datamap.nDstCh = nAddPhoneCh;

   if (IFX_TRUE == fDoMapping)
   {
      /* Map phone channel to data channel */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Map phone ch %d to data ch %d\n",
           (int)nAddPhoneCh, (int)nDataCh));

      ret = ioctl(fd_ch, IFX_TAPI_MAP_DATA_ADD, (IFX_int32_t) &datamap);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error mapping phone ch %d to data ch %d using fd %d. \
              (File: %s, line: %d)\n",
              (int) nAddPhoneCh, (int) nDataCh, (int) fd_ch,
               __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }
   else
   {
      /* Unmap phone channel to data channel */
      TRACE(TAPIDEMO, DBG_LEVEL_LOW, ("Unmap phone ch %d from data ch %d\n",
           (int)nAddPhoneCh, (int)nDataCh));

      ret = ioctl(fd_ch, IFX_TAPI_MAP_DATA_REMOVE, (IFX_int32_t) &datamap);
      if (IFX_ERROR == ret)
      {
         TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
              ("Error unmapping phone ch %d from data ch %d using fd %d. \
              (File: %s, line: %d)\n",
              (int) nAddPhoneCh, (int) nDataCh, (int) fd_ch,
               __FILE__, __LINE__));
         return IFX_ERROR;
      }
   }

   return ret;
} /* VOIP_MapPhoneToData() */


/**
   Initialize socket for incoming voice packets

   \param pProgramArg - pointer to program arguments
   \param nChIdx - channel index
   \param nPort  - which port is socket using

   \return socket number or NO_SOCKET if error.
*/
IFX_int32_t VOIP_InitUdpSocket(PROGRAM_ARG_t* pProgramArg,
                               IFX_int32_t nChIdx,
                               IFX_int32_t* nPort)
{
   IFX_int32_t s;
   struct sockaddr_in my_addr;

   /* check input parameters */
   TD_PTR_CHECK(pProgramArg, NO_SOCKET);
   TD_PTR_CHECK(nPort, NO_SOCKET);
   TD_PARAMETER_CHECK((0 > nChIdx), nChIdx, NO_SOCKET);

   s = socket(PF_INET, SOCK_DGRAM, 0);
   if (NO_SOCKET == s)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("InitUdpSocket: Can't create UDP socket, %s. \
           (File: %s, line: %d)\n",
            strerror(errno), __FILE__, __LINE__));
      return NO_SOCKET;
   }

   bzero((IFX_char_t*) &my_addr, (IFX_int32_t) sizeof(my_addr));
   my_addr.sin_family = AF_INET;

/*   my_addr.sin_port = htons(VOICE_UDP_PORT + nChIdx + nFreeSocketPortAvail);*/

   /* According to the standard it is recomended to send RTP packets on even
    ports while odd ones are reserved for RTCP */

   my_addr.sin_port = htons(VOICE_UDP_PORT + 2 * nChIdx + nFreeSocketPortAvail);

   if (pProgramArg->oMy_IP_Addr.sin_addr.s_addr)
   {
      my_addr.sin_addr.s_addr = pProgramArg->oMy_IP_Addr.sin_addr.s_addr;
   }
   else
   {
      my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   }

   if (-1 == (bind(s, (struct sockaddr *) &my_addr, sizeof(struct sockaddr))))
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("InitUdpSocket: Can't bind to port, %s. (File: %s, line: %d)\n",
            strerror(errno), __FILE__, __LINE__));
      close(s);
      return NO_SOCKET;
   }

#ifndef VXWORKS
   /* make the socket non blocking */
   fcntl(s, F_SETFL, O_NONBLOCK);
#endif /* VXWORKS */

   *nPort = my_addr.sin_port;

   return s;
} /* VOIP_InitUdpSocket() */


/**
   Handles data from socket

   \param pPhone - pointer to PHONE_t
   \param pConn  - pointer to phone connection
   \param fQosHandlesData - if IFX_TRUE QoS is handling data,
                            if IFX_FALSE we handle data.

   \return IFX_SUCCESS on ok otherwise IFX_ERROR
*/
IFX_return_t VOIP_HandleSocketData(PHONE_t* pPhone,
                                   CONNECTION_t* pConn,
                                   IFX_boolean_t fQosHandlesData)
{
   /* buffer size increased because packets exceed 400 B
      for some codecs with bigger packetization lengths */
   static IFX_char_t buf[MAX_PACKET_SIZE];
   IFX_int32_t size = 0;
   IFX_return_t ret = IFX_SUCCESS;
   struct sockaddr_in oToAddr;

   /* check input parameters */
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);
   TD_PARAMETER_CHECK((0 >= pConn->nUsedSocket), pConn->nUsedSocket, IFX_ERROR);

   /* get address structure size */
   size = sizeof(pConn->oConnPeer.oRemote.oToAddr);
   /* get data from socket */
   ret = recvfrom(pConn->nUsedSocket,
                  buf, sizeof(buf), 0,
                  (struct sockaddr *) &oToAddr,
                  /* size is set by recvfrom to size of address structure(arg5)
                     used by recvfrom() */
                  (int *) (IFX_int32_t) &size);
   /* aditional check if peer address is the same as packet from address */
   if (pConn->oConnPeer.oRemote.oToAddr.sin_addr.s_addr != 
      oToAddr.sin_addr.s_addr ||
      pConn->oConnPeer.oRemote.oToAddr.sin_port != oToAddr.sin_port)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error, Received from %s:%d\n",
             inet_ntoa(oToAddr.sin_addr), oToAddr.sin_port));
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error, Should be from %s:%d\n",
             inet_ntoa(pConn->oConnPeer.oRemote.oToAddr.sin_addr),
             pConn->oConnPeer.oRemote.oToAddr.sin_port));
   }
   /* check if recvfrom() error */
   if (0 > ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
            ("Error reading from socket %d, %s. (File: %s, line: %d)\n",
            (IFX_int32_t) pConn->nUsedSocket, strerror(errno), 
            __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /* check if recvfrom() read data */
   else if (0 == ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
            ("No data read from socket %d.\n",
             (IFX_int32_t) pConn->nUsedSocket));
      return IFX_ERROR;
   }
   /* data read handle it */
   else
   {
      /* check if QoS handles data */
      if (IFX_TRUE == fQosHandlesData)
      {
         /* UDP redirection is handling voice streaming */
         return IFX_SUCCESS;
      }
      /* send packets only when connection in active state */
      if (TD_GET_STATE_INFO(pPhone->rgStateMachine[FXS_SM].nState,
                            TD_STATE_DATA_TRANSMIT))
      {
         /* application handles the data not QoS */
         ret = write(pConn->nUsedCh_FD, buf, ret);
         /* check if write successfull */
         if (0 > ret)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                 ("Error writing data to device, %s. (File: %s, line: %d)\n",
                  strerror(errno), __FILE__, __LINE__));
            return IFX_ERROR;
         }
         else if (0 == ret)
         {
            TRACE(TAPIDEMO, DBG_LEVEL_NORMAL, ("No data written to device.\n"));
            return IFX_ERROR;
         }
      } /* if right state */
   } /* if ( 0 >/=/< ret ) */
   return IFX_SUCCESS;
} /* VOIP_HandleSocketData() */


/**
   Handles data from chip

   \param pCtrl - pointer to status control structure
   \param pPhone  - pointer to PHONE
   \param nDataCh_FD - data channel file descriptor
   \param pConn - pointer to phone connection
   \param fQosHandlesData - if IFX_TRUE QoS is handling data,
                            if IFX_FALSE we handle data.

   \return status (IFX_SUCCESS or IFX_ERROR)
*/
IFX_return_t VOIP_HandleData(CTRL_STATUS_t* pCtrl,
                             PHONE_t* pPhone,
                             IFX_int32_t nDataCh_FD,
                             CONNECTION_t* pConn,
                             IFX_boolean_t fQosHandlesData)
{
   IFX_return_t ret = IFX_SUCCESS;
   IFX_int32_t len = 0;
   /* buffer size increased because packets exceed 400 B
      for some codecs with bigger packetization lengths */
   static IFX_char_t buf[MAX_PACKET_SIZE];

   /* check input parameters */
   TD_PTR_CHECK(pCtrl, IFX_ERROR);
   TD_PTR_CHECK(pPhone, IFX_ERROR);
   TD_PTR_CHECK(pConn, IFX_ERROR);

   /* Get data from driver - DTMF, talking on phone */
   ret = read(nDataCh_FD, buf, sizeof(buf));
   /* check if no error occured during read */
   if (0 > ret)
   {
      TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
           ("No data read from device, fd %d. (File: %s, line: %d)\n",
            (int) nDataCh_FD, __FILE__, __LINE__));
      return IFX_ERROR;
   }
   /* check if QoS handles data */
   if (IFX_TRUE == fQosHandlesData)
   {
      /* Don`t use application to handle the data, QoS is used */
      return IFX_SUCCESS;
   }
   /* check if data was read */
   if (0 < ret)
   {
      /* send packets only when connection in active state */
      if (TD_GET_STATE_INFO(pPhone->rgStateMachine[FXS_SM].nState,
                            TD_STATE_DATA_TRANSMIT))
      {
         /* check connection type */
         if (LOCAL_CALL == pConn->nType ||
             LOCAL_BOARD_CALL == pConn->nType)
         {
            /* local call send packet to data channel not supported for 
             * conference */
            if (NO_CONFERENCE == pPhone->nConfIdx)
            {
               /* Local call - write to data channel */
               len = write(pConn->oConnPeer.oLocal.pPhone->nDataCh_FD, buf, 
                  ret);
               /* check if write successfull */
               if (0 > len)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                        ("Error, Error writing data to device %d, %s. "
                         "(File: %s, line: %d)\n",
                         (int) pConn->oConnPeer.oLocal.pPhone->nDataCh_FD,
                         strerror(errno), __FILE__, __LINE__));
                  return IFX_ERROR;
               }
               else if (0 == len)
               {
                  TRACE(TAPIDEMO, DBG_LEVEL_NORMAL,
                        ("No data written to device %d.\n",
                         (int) pConn->oConnPeer.oLocal.pPhone->nDataCh_FD));
                  return IFX_ERROR;
               }
            } /* NO_CONFERENCE == pPhone->nConfIdx */
         }
         else if (EXTERN_VOIP_CALL == pConn->nType)
         {
            /* External call send data through socket */
            len = sendto(pConn->nUsedSocket, buf, ret, 0,
                         (struct sockaddr *) &pConn->oConnPeer.oRemote.oToAddr,
                         sizeof(pConn->oConnPeer.oRemote.oToAddr));
            if (ret != len)
            {
               TRACE(TAPIDEMO, DBG_LEVEL_HIGH,
                     ("Error calling sendto(), %s, socket %d, data ch %d."
                      " (File: %s, line: %d)\n",
                      strerror(errno), (int) pConn->nUsedSocket,
                      (int) pConn->nUsedCh, __FILE__, __LINE__));
               return IFX_ERROR;
            }
         } /* if (... == pConn->nType) */
      } /* if in active state state */
   } /* if data was read */

   return IFX_SUCCESS;
} /* VOIP_HandleData() */

/**
   Reads data from file descriptor (can be data channel or socket) to empty it.

   \param nFd  - file descriptor to read from

   \return status (IFX_SUCCESS or IFX_ERROR)
*/
IFX_return_t VOIP_DiscardData(IFX_int32_t nFd)
{
   static IFX_char_t buf[MAX_PACKET_SIZE];

   /* check input parameters */
   TD_PARAMETER_CHECK((0 >= nFd), nFd, IFX_ERROR);

   /* read data from fd */
   if (0 > read(nFd, buf, sizeof(buf)))
   {
      return IFX_ERROR;
   }

   return IFX_SUCCESS;
}

/**
   Get file descriptor of device for data channel.

   \param nDataCh - data channel
   \param pBoard - board handle

   \return device connected to this channel or NO_DEVICE_FD if none
*/
IFX_int32_t VOIP_GetFD_OfCh(IFX_int32_t nDataCh, BOARD_t* pBoard)
{
   /* check input parameters */
   TD_PTR_CHECK(pBoard, NO_DEVICE_FD);
   TD_PARAMETER_CHECK((0 > nDataCh), nDataCh, NO_DEVICE_FD);
   TD_PARAMETER_CHECK((pBoard->nMaxCoderCh <= nDataCh), nDataCh, NO_DEVICE_FD);

   return Common_GetDeviceOfCh(pBoard, nDataCh);
} /* VOIP_GetFD_OfCh() */
